#pragma warning( disable : 4786)
#ifndef HPP_LDRAW
#define HPP_LDRAW

#include <iostream>
#include <fstream>
#include <cstdlib>
#include <math.h>
#include <cstring>
#include <string.h>
#include <vector>
#include <algorithm>

using namespace std;

#define Comment 0
#define Subfile 1
#define Line 2
#define Triangle 3
#define Quad 4
#define Condline 5
#define Empty 6

#define EPSILON 0.001
#define UNIFYDIST 0.005
#define PI acos(-1)

#define FLAT 0.1; // Small angle, no line generated
#define MAXCOND 90; // Condline generated below this angle
#define MINEDGE 60; // Edge line generated above this angle

// Global variables

#ifdef MAIN
double LdPrecision = EPSILON;
double UnifyDist = UNIFYDIST;
double LdFlat = FLAT;
double LdMaxCond = MAXCOND;
double LdMinEdge = MINEDGE;
bool ManageCond = true;
bool DelEdge = false;
bool ExpandPattern = false;
bool ExpandFormer = false;
bool Inlined = true;
bool OnlyUnmatched=false;
bool NoUnmatched=false;
bool NoReTriangulate=false;
bool SubComments = false;
bool NoProject=false;
bool Merge=false;
bool DebugColors = false;
bool ExpandRect=true;
bool ExpandQuads=true;
bool Uncolor=false;

int CxCv=0;
int CompPoint=0;
int KeyColor=-1;

bool xdir=false;
bool zdir=false;

#else
extern double LdPrecision;
extern double UnifyDist;
extern double LdFlat;
extern double LdMaxCond;
extern double LdMinEdge;
extern bool ManageCond;
extern bool DelEdge;
extern bool ExpandPattern;
extern bool ExpandFormer;
extern bool Merge;
extern bool Inlined;
extern bool NoProject;
extern bool DebugColors;
extern int CxCv;
extern bool OnlyUnmatched;
extern bool NoUnmatched;
extern bool NoReTriangulate;
extern bool SubComments;
extern bool ExpandRect;
extern bool ExpandQuads;
extern int CompPoint;
extern int KeyColor;
extern bool Uncolor;

extern bool xdir;
extern bool zdir;

#endif

class Object
{
public:
	virtual void printclassname()	{}

// swap x and y coordinates
	virtual void swapXY() {}

// swap y and z coordinates
	virtual void swapYZ() {}

// Flatten object in Y direction
	virtual void flattenY() {}

};

class Point:public Object
{
public:

	// constructeurs
	Point();
	Point(const Point &p);
	Point(double x, double y, double z );

	//methodes
	double getx() { return m_x; }
	double gety() { return m_y; }
	double getz() { return m_z; }
	void swapXY();
	void swapYZ();

	void setp (double xi, double yi, double zi);
	void setp (const Point &p);
	Point::addp(Point* p);
	char* pparse (char* str);
	double distance (Point &p=Point(0,0,0));
	bool compare(Point &p);
	void print(bool lf=true);
	string toString();
	void printclassname();
	void flattenY() { m_y = 0;}
	double getYOrientation(Point &p1, Point &p3);

	// return signed distance to line p1p2
	double getLineDist(Point &p1, Point &p2);

protected: // == Private except for derived class
	double m_x,m_y,m_z;

};

// Center of mass of a collection of points
class PointBary : public Point
{
public:
	PointBary(Point &p);
	void addPoint(Point &p);
	void getPoint(Point &p);

private:
	double m_xsum, m_ysum, m_zsum;
	int nbpoint;
};


// collect nearby points into their center of mass
class PointUnify : public Object
{
public:
	~PointUnify();
	void addPoint(Point &p);
	void getPoint(Point &p);
	string toString(); 

private:
	vector <PointBary*> barytab;
};

class Vector:public Point
{
public:
	
	// Vector derived from Point
	Vector(Point &p):Point(p)
	{
	}
	// Vector derived from Point
	Vector(double x, double y, double z ):Point(x,y,z)
	{
	}
	Vector(Point &p1, Point &p2);

	Vector* cross (Vector &v);
	double dot (Vector &v);
	double norm ();
	void normalize ();
	// perp product in x-z plane
	double perp(Vector &v);
	void mult(double k);
};	


class Matrix:public Object
{
public:
	Matrix::Matrix();
	Matrix(Point &p1, Point &p2, Point &p3);
	Matrix(double a,double b,double c,double d,double e,double f,double g,double h,double i);
	string toString();
	Vector* Mult(Vector &v1);
	Matrix* MatMult(Matrix &m1);
	double MatDeterminant();
	void flattenY(void);
	void projectY(Vector *normal);
	void swapXY();
	void swapYZ();
protected:
	double m_mat[3][3];
};

class LdObject:public Object
{
public:
	LdObject()
	{
		m_NoOut=false;	
	}
	virtual string toString()
	{
		return "";
	}

	virtual int getType()
	{
		return -1;
	}

// return a pointer to the first line of an object
	virtual LdObject* getFirstLine()
	{
		return NULL;
	}
	
// return a pointer to the first line of an object
	virtual LdObject* getNextLine()
	{
		return NULL;
	}
	
// Get orientation (the object is supposed in the X-Z plane
	virtual double getYOrientation()
	{
		return 1;
	}

// Test if all vertices of an object are on the outside of an oriented line.
	virtual bool isOutside(LdObject* line)
	{
		return true;
	}

// Test if both ends of a segment are outside a surface
	virtual bool SegmentIsOutside(LdObject* line)
	{
		return false;
	}
// Test for degenerated elements (identical vertices)
	virtual bool isInvalid(LdObject* &ldo)
	{
		return false;
	}

// add points to point unify structure
	virtual void addPoints (PointUnify &PU) {return;}
// retrieve points from point unify structure
	virtual void getPoints (PointUnify &PU) {return;}

// Project a point on a surface in Y direction
	virtual bool projectPointY(Point &p) {return false;}

// Project a surface or line on surface surf in Y direction
	virtual void projectY(LdObject *surf) {return;}

// Return true if surfaces overlap
	bool Overlap(LdObject* surf);

	void SetNoOut(bool b) { m_NoOut=b; }
	bool GetNoOut() {	return m_NoOut; }

	void SetBFC(int i) {m_BFC=i;}
	int GetBFC() {return m_BFC;}

// return true if triangle or quad
	virtual bool isSurface() {return false;}
// return true if line or condline
	virtual bool isLine() {return false;}

// Translate to point p
	virtual translate(Point &p)
	{
	}
// Transform with matrix m
	virtual transform(Matrix &m)
	{
	}	

// return normal to a surface
	virtual Vector* getNormal()
	{
		return NULL;
	}

// return angle between two surfaces
	virtual double getAngle(LdObject &surf, LdObject &line)
	{
		return 0L;
	}

// return true if point matches one line end
	virtual bool matchPoint(Point &p1) { return false;}

// Change winding of a surface
	virtual void InvertWinding()
	{
		return;
	}

// Set color of element
	virtual void inline SetColor (int col)
	{
	}

// Get color of element
	virtual inline int  getColor (void)
	{
		return 0;
	}

// return true if line found
	virtual bool matchLine(LdObject &line) {return false;}

// return surface vertex not end of line
	virtual Point* getOtherPoint (LdObject &line) {return NULL;}

// if two consecutive points of surface matches p1 and p2, 
//	copies of one or two remaining points are returned
// if no match, *p1==NULL
	virtual void getOtherPoints(Point* p1, Point* p2) {return;}

// Cut object with oriented line,  If points defining lines are outside object,
// no cut is done. Only objects "inside" line are returned.
	virtual void Cut(LdObject* line, LdObject* &o1, LdObject* &o2)
	{
		o1=NULL;
		o2=NULL;
	}

// Rotate surface points
	virtual void RotPoints(void)
	{
		return;
	}
// Returns a pointer to next point of a surface
	virtual Point* getNextPoint(void)
	{
		return NULL;
	}

private:
	bool m_NoOut;
	int m_BFC;
};


struct Polyvertex
{
	Point *p;
	double azimuth;
};

class Poly:public Object
{
public:
	// add new vertex to poly. Check if really new before adding.
	void addNewVertex(Point* p);
	// leftmost point (lowest x) is placed first in list.
	void getBottomMost(void);
	// calculate cotg of elevation angle (x/z)
	void setAzimuth(void);
	// sort according to azimuth
	void sortAz(void);
	void getConvexHull(void);
	// retrieve one triangle or quad. NULL if nothing left
	LdObject* getLdSurf(int color);

	string toString(void);

private:
	vector <Polyvertex*> Vertices;
};


class LdComment:public LdObject
{
public:
	LdComment(const LdComment &ldc);
	LdComment(const string &str);
	LdComment(const char *str);
	
	string toString();
	inline int getType()
	{
		return Comment;
	}

	int readbfc();

private:
	string m_cs;
};

class LdSubfile:public LdObject
{
public:
	bool isOutside(LdObject* line);
	LdSubfile(const LdSubfile &lds);
	LdSubfile(const char *str);
	translate(Point &p);
	transform(Matrix &m);
	void flattenY(void);
	Point * getOrigin() {return &m_origin;}
	string getName() {return m_file;}
	Matrix * getMAtrix() {return &m_mat;}
	void swapXY();
	void swapYZ();
	void projectY(LdObject *surf);
	string toString();
	inline int getType()
	{
		return Subfile;
	}
	inline void SetColor (int col)
	{
		m_col = col;
	}
	int inline getColor (void)
	{
		return m_col;
	}
	

private:
	int m_col;
	Point m_origin;
	Matrix m_mat;
	string m_file;
};

class LdLine:public LdObject
{
public:
	LdLine(const LdLine &l);
	LdLine(const char *str);
	LdLine(const Point &p1, const Point &p2, const int col=0);
	string toString();
	inline int getType()
	{
		return Line;
	}

	inline void SetColor (int col)
	{
		m_col = col;
	}
	int inline getColor (void)
	{
		return m_col;
	}
	
	inline SetLinePoints (Point p1, Point p2)
	{
		m_p1 = p1;
		m_p2 = p2;
	}
	void swapXY();
	void swapYZ();
	translate(Point &p);
	transform(Matrix &m);
	LdObject* getFirstLine();
	bool isOutside(LdObject* line);
	inline bool isLine() {return true;}
	bool matchLine(LdObject &line);
	Point * getP1() {return &m_p1;}
	Point * getP2() {return &m_p2;}
	bool matchPoint(Point &p1);
	void projectY(LdObject *surf);
	void InvertWinding();
	bool isInvalid(LdObject* &ldo);
	void flattenY();
	void addPoints (PointUnify &PU);
	void getPoints (PointUnify &PU);
	// Intersection segment-line
	bool inter (LdLine *line, Point *p);
	bool inter (Point* p1, Point* p2, Point *p);
	void Cut(LdObject* line, LdObject* &o1, LdObject* &o2);
	
private:
	int m_col;
	Point m_p1;
	Point m_p2;
	double dist[2];
};

class LdTriangle:public LdObject
{
public:
	LdTriangle(const LdTriangle &t);
	LdTriangle(const char *str);
	LdTriangle(Point &p1, Point &p2, Point &p3, int col);
	string toString();
	inline int getType()
	{
		return Triangle;
	}
	inline void SetColor (int col)
	{
		m_col = col;
	}
	int inline getColor (void)
	{
		return m_col;
	}
	
	void swapXY();
	void swapYZ();
	void RotPoints(void);
	translate(Point &p);
	transform(Matrix &m);
	LdObject* getFirstLine();
	LdObject* getNextLine();
	// Project point p on triangle in Y direction
	bool projectPointY(Point &p);
	void projectY(LdObject *surf);
	bool isOutside(LdObject* line);
	bool SegmentIsOutside(LdObject* line);
	inline bool isSurface() {return true;}
	bool matchLine(LdObject &line);
	Vector* getNormal();
	double getAngle(LdObject &surf, LdObject &line);
	double getAngle(LdObject &surf);
	Point* getOtherPoint (LdObject &line);
	void getOtherPoints(Point* p1, Point* p2);
	void InvertWinding();
	bool isInvalid(LdObject* &ldo);
	void flattenY();
	void addPoints (PointUnify &PU);
	void getPoints (PointUnify &PU);
	double getYOrientation();
	void Cut(LdObject* line, LdObject* &o1, LdObject* &o2);

private:
	int m_col;
	Point m_p1;
	Point m_p2;
	Point m_p3;
	int m_edgecpt;
	double dist[3];
};

class LdQuad:public LdObject
{
public:
	LdQuad(const char *str);
	LdQuad(const LdQuad &q);
	LdQuad(Point &p1, Point &p2, Point &p3, Point &p4, int col);

	string toString();
	inline int getType()
	{
		return Quad;
	}
	inline void SetColor (int col)
	{
		m_col = col;
	}
	int inline getColor (void)
	{
		return m_col;
	}
	
	void swapXY();
	void swapYZ();
	void RotPoints(void);
	translate(Point &p);
	transform(Matrix &m);
	LdObject* getFirstLine();
	LdObject* getNextLine();
	bool projectPointY(Point &p);
	void projectY(LdObject *surf); 
	bool isOutside(LdObject* line);
	inline bool isSurface() {return true;}
	bool matchLine(LdObject &line);
	Vector* getNormal();
	double getAngle(LdObject &surf, LdObject &line);
	Point* getOtherPoint (LdObject &line);
	void getOtherPoints(Point* p1, Point* p2);
	void InvertWinding();
	bool isInvalid(LdObject* &ldo);
	void flattenY();
	void addPoints (PointUnify &PU);
	void getPoints (PointUnify &PU);
	double getYOrientation();
	void Cut(LdObject* line, LdObject* &o1, LdObject* &o2);
	void SplitQuad(LdObject* &triangle1, LdObject* &triangle2, LdObject* &condline);

private:
	int m_col;
	Point m_p1;
	Point m_p2;
	Point m_p3;
	Point m_p4;
	double dist[4];
	int m_edgecpt;
};

class LdCondline:public LdObject
{
public:
	LdCondline(const char *str);
	LdCondline(Point &p1, Point &p2, Point &p3, Point &p4, const int col=0);
	string toString();
	inline int getType()
	{
		return Condline;
	}
	inline void SetColor (int col)
	{
		m_col = col;
	}
	int inline getColor (void)
	{
		return m_col;
	}
	void swapXY();
	void swapYZ();
	translate(Point &p);
	transform(Matrix &m);
	void addPoints (PointUnify &PU);
	void getPoints (PointUnify &PU);
	LdObject* getFirstLine();
	inline bool isLine() {return true;}
	bool matchLine(LdObject &line);

private:
	int m_col;
	Point m_p1;
	Point m_p2;
	Point m_p3;
	Point m_p4;
};

class LdEmpty:public LdObject
{
	string toString()
	{
		return "";
	}
	int getType()
	{
		return Empty;
	}
};

static void funcdelete (LdObject* ldo); 


class LdPath
{
public:
	LdPath();
	~LdPath();
	addPath(string &s);
	bool subOpen(ifstream &ldSubfile, string &subfilename);
	string toString();
private:
	vector <string*> pathtable;
};


class Fldraw:public LdObject
{

public:

	Fldraw()
	{
	}

	Fldraw(Fldraw &fld);

	~Fldraw();
	

	string toString() { return NULL; }
	inline int getType()
	{
		return -1;
	}


// Method to load ldraw file
	bool ldLoad (char *fname);
// Method to save ldraw file
	bool ldSave (char *fname);
// Expand subfiles
	ldExpand(LdPath *path);
// Swap coordinates
	void swapXY();
	void swapYZ();

// Force Y orientation and flatten
	void ForceYOrientation(void);

// Invert winding of tris/quads
	void Fldraw::InvertWinding(void);

// Method to scan array and insert lines on non-matching edges
	findOrphan();
// Method to scan array and insert lines/condlines on adjacent surfaces
	findAdjacent();

// Regenerate condlines surrounding surface t1
	void GenCondlines(LdObject* t1);
// Delete condlines adjacent to surface t1
	void RemCondlines(LdObject* t1);	
// Uncolor all elements of a LDraw file
	unColor();
// test functions
	void Test(LdObject* surf);
	void Test2(Fldraw* fldo);

// Copy rferences to all elements. Elements must not be deleted elsewhere !
	void LdCopy(Fldraw* source);

// Return a fldraw object containing all overlapping elements
	Fldraw* getOverlapping(LdObject* surf);

// cuts all elements ar the edges of surf.
	void CutAtEdges(LdObject* surf);

// Retriangulate elements in the object. 
// Assumes convexity and maximum polygon size of 8 (intersection of 2 diagonal quads)
	void Retriangulate();

// cuts all elements at the edges of surfaces of "form" then projects on equivallent form3D
	Fldraw* CutAll(Fldraw &form, Fldraw &form3D);

// Split quads into triangles. Act only on key-colored quads (all if keycolor == -1)
	void SplitQuad(void);

// Copy condlines 
	void CopyCond(Fldraw &source);

// Unify vertices
	void Unify(void);

	void printptrs(void)
	{
		cout << endl << ldtab.size() << endl;
		for(int i=0; i< ldtab.size(); i++) 
		{
			cout << ldtab[i] << " ";
			cout << ldtab[i]->toString().c_str() << endl;
		}
	}



private:
	// Ldraw object array of pointers
//	LdObject **ldtab;
	vector <LdObject*> ldtab;
	unsigned int m_nLine; // Initial line number
//	int m_nextLine; // Insertion point in array
//	int m_maxLine; // Array end;
};


#endif
