#include <iostream>
#include <fstream>
#include <cstdlib>
#include <math.h>
#include <time.h>
#include <time.h>

using namespace std;

#define MAIN
#include "ldraw.hpp"

void pause(void)
{
	cout << "Press <Enter> to quit" << endl;
	getchar();
}
void usage(void)
{
	cout << "Unrectifier [options] InFile Outfile" << endl;
	cout << "-k <val>: key color (default none)" << endl;
	cout << "-r: Don't expand Rect primitives" << endl;
	cout << "-q: Don't split quads" << endl;
	cout << "-c: Don't manage condlines around split quads" << endl;
	cout << "-u: Return file to uncolored state" << endl;
	cout << "-l \"path\": define ldraw search path" << endl;
	pause();
}

int main(int argc, char *argv[])
{
	printf("\nUnrectifier 1.0 - by Philo");
	printf("\n--------------------------\n");

	if(argc<3) 
	{
		usage();
		return 1;
	}
	bool UnifyOutput=true;

	LdPath* path=new LdPath();
	// Get input file path
	string curpath=argv[argc-2];
	size_t pathlen = curpath.find_last_of("\\");
	if(pathlen == string::npos) curpath = ".\\"; else curpath.resize(pathlen+1);
	path->addPath(curpath);
	string derivedpath=curpath;
	path->addPath(derivedpath.append("p\\"));
	derivedpath=curpath;
	path->addPath(derivedpath.append("s\\"));

	for(int i=1; i<argc-2; i++)
	{
		if(argv[i][0] != '-')
		{
			usage();
			return 1;
		}
		switch (argv[i][1])
		{
		case 'k':
			i++;
			sscanf(argv[i], "%d",&KeyColor);
			break;
		case 'r':
			ExpandRect=false;
			Inlined = false;
			break;
		case 'q':
			ExpandQuads=false;
			break;
		case 'c':
			ManageCond=false;
			break;
		case 'u':
			Uncolor=true;
			break;
		case 's':
			switch (argv[i][2])
			{
			case 'f':
				ExpandFormer=true;
				break;
			case 'p':
				ExpandPattern=true;
				break;
			default:
				usage();
				return 1;
			}
			break;

		case 'l':
			i++;
			curpath = argv[i];
			if(curpath.at(curpath.size()-1) != '\\') 
			{
				curpath.append("\\");
			}
			derivedpath=curpath;
			path->addPath(derivedpath.append("p\\"));
			derivedpath=curpath;
			path->addPath(derivedpath.append("parts\\"));
			path->addPath(derivedpath.append("s\\"));
			curpath.append("Unofficial\\");
			derivedpath=curpath;
			path->addPath(derivedpath.append("p\\"));
			derivedpath=curpath;
			path->addPath(derivedpath.append("parts\\"));
			path->addPath(derivedpath.append("s\\"));

			break;
		default:
			usage();
			return 1;
		}
	}

	Fldraw LDfile;
	cout << endl << "Reading input file " << argv[argc-2] << endl;
	if(!LDfile.ldLoad (argv[argc-2]))
	{
		cout << argv[argc-2] << " not found" << endl;
		pause();
		return 2;
	}

	if(ExpandRect)
	{
		cout << endl << "LDraw subfiles search locations:\n" << path->toString().c_str() << endl;
		if(KeyColor<0) cout << "\nExpanding Rect primitives\n";
		else  cout << "\nExpanding key-colored primitives\n";
		LDfile.ldExpand(path);
		cout << endl;
	}
	if(ExpandQuads)
	{
		cout << "\nSplitting quads\n";
		LDfile.SplitQuad();
		cout << endl;
	}
	

	if(Uncolor)
	{
		cout << endl << "Return file to uncolored state" << endl;
		LDfile.unColor();
	}
	cout << endl << "Writing output file " << argv[argc-1] << endl;
	if(!LDfile.ldSave(argv[argc-1]))
	{
		cout <<" Can't create" << argv[argc-1] << endl;
		pause();
		return 2;
	}
	pause();
	return 0;
}

